/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * CUNIX.C  Serial communications routines for Unix (POSIX) systems
 */
 
 /*
  * Based on the Unix communications used by J. Hadi Salim in the
  * Linux port of `casio' communications software for the data diary.
  *
  * Information also obtained from a WWW page by <mumble> at:
  *   http://www.easysw.com/~mike/serial/
  */

#define _BSD_SOURCE /* to allow usleep() etc in Linux */

#include "cas.h"
#include "misc.h"
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include "commsys.h"

#define COMM_TIMEOUT          50000L     /* 50,000 */

static int portfd=0;
static int portflags;
static int oldflags;
static struct termios oldterm;
static struct termios portterm;

int COMsysInit(char *filename, int baud, char parity, int handshake)
{
  int parnum=PARENB;
  int hand=0;
  int baudrate=B9600;

  switch (parity) {
    case CASPORT_NONE: parnum=0;break;
    case CASPORT_ODD:  parnum=PARENB|PARODD;break;
    case CASPORT_EVEN:
    default:           parnum=PARENB;break;
  }

  switch (baud) {
    case 1200: baudrate=B1200;break;
    case 2400: baudrate=B2400;break;
    case 4800: baudrate=B4800;break;
    case 9600: baudrate=B9600;break;
    default:
    Msg(EXIT_COMM,"comm","Unsupported baud rate `%d'",baud);
  }    
  
  /* not used at present */
  if (handshake&CASPORT_DTR) hand |=TIOCM_DTR;
  if (handshake&CASPORT_RTS) hand |=TIOCM_RTS;

  /* Initialise COM port */

  if ((portfd=open(filename, O_RDWR | O_NOCTTY | O_NDELAY)) < 0)
    Msg(EXIT_COMM,"comm","Unable to open port `%s'",filename);

  atexit(COMsysClose);  /* cleans up COM port upon exit */

  if (!isatty(portfd))
    Msg(EXIT_COMM,"comm","`%s' is not a tty",filename);
                        
  /* get port attributes, store in oldterm */
  if (tcgetattr(portfd, &oldterm) < 0)
    Msg(EXIT_COMM,"comm","Unable to read current port attributes for `%s'",filename);
  if ((portflags=oldflags=fcntl(portfd,F_GETFL))<0)
    Msg(EXIT_COMM,"comm","Unable to read flags for `%s'",filename);
  
  /* copy old attrs into new structure, so we can restore tty */
  portterm=oldterm;
  
  portterm.c_cflag &= ~(PARENB | PARODD);
  portterm.c_cflag |= parnum;
  portterm.c_cflag &= ~CSIZE;
  portterm.c_cflag |= CREAD | CS8 | CLOCAL;
  cfsetispeed (&portterm, baudrate);
  cfsetospeed (&portterm, baudrate);

  /* set non-canonical mode */
  portterm.c_lflag = 0;
    
  /* Ignore breaks and make terminal raw and dumb. */
  portterm.c_iflag = 0;
  portterm.c_iflag |= IGNBRK;
  portterm.c_oflag &= ~OPOST;

  /* now clean the serial line and activate the settings for casio */
  tcflush(portfd, TCIOFLUSH);
  if (tcsetattr(portfd, TCSANOW, &portterm) < 0)
    Msg(EXIT_COMM,"comm","Unable to set terminal settings");

  /* set non-blocking */
  fcntl(portfd,F_SETFL,(portflags |= O_NONBLOCK));

  /* wait a little while for everything to settle through */
  
  usleep(3000);
  
  return TRUE;
}


void COMsysClose(void)
{
  if (!portfd) /* already closed */
    return;
  
  /* wait a while to let calc receive everything.
     tcsetattr doesn't seem to wait :( */
  usleep(3000);
    
  if (tcsetattr(portfd, TCSADRAIN, &oldterm) < 0)
    Msg(M_WARN,"comm/close","Unable to reset serial port");
  fcntl(portfd,F_SETFL,oldflags);
  
  close(portfd);
  portfd=0;
}


void COMsysPutChar(unsigned char data)
{
  if (write (portfd, &data, 1) != 1)
    Msg(EXIT_COMM,"comm/write","Unable to write char 0x%02x",data);
#ifdef DEBUG
  printf("(%02x)",data);fflush(stdout);
#endif
}


unsigned char COMsysGetChar(void)
{
  unsigned char rdchar;
  int ok;
  
  do {
    ok=read(portfd, &rdchar, 1);
    if (ok==0) Msg(M_DEBUG,"comm/read","EOF encountered");
  } while (ok <= 0);
#ifdef DEBUG
  printf("[%02x]",rdchar);fflush(stdout);
#endif
  
  return rdchar;
}


unsigned char COMsysGetCharTimeout(int *success_r,unsigned int modifier)
{
  unsigned long x;
  unsigned long timeout;
  int ok;
  unsigned char rdchar;

  timeout=COMM_TIMEOUT * 100 / modifier;
  *success_r=TRUE;
  
  for(x=0;x<timeout;x++) {
    ok=read(portfd, &rdchar, 1);
    if (ok>0) {
#ifdef DEBUG
      printf("{%02x}",rdchar);fflush(stdout);
#endif
      return rdchar & 0xff;
    }
  }

  *success_r=FALSE;
  return 1;
}


/* end of cunix.c */
