/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * FCOM.C  Routines for serial communications.  Calls functions in sXXXX.c
 * with pointers to functions of the appropriate type in cXXXXX.c
 */
 
#define _CAS_FCOM_C_

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "s9700.h"
#include "s9850.h"
#include "sraw.h"
#include "fcom.h"
#include "commsys.h"

#ifdef CAS_COMM /* covers most of this file */ 

void COMReadData(char *filename, casinfo *info)
{
  darg_t dargs;
  int baud,parity,hand;
  unsigned char rdchar;
  void *hdr=NULL;

  args_join(flags.input.args,ARGS_DEF_IN,fCOM,&dargs);
  COMParseArgs(dargs, &baud, &parity, &hand);

  flg_bytes=args_is_set(dargs,"bytes",NULL);

  Msg(M_VERBOSE,"com","Settings: %s, %u, 8, %c, 2%s%s",filename,baud,parity,
    hand&CASPORT_DTR?", DTR":"",hand&CASPORT_RTS?", RTS":"");

  if (!(COMsysInit(filename,baud,parity,hand))) {
    Msg(M_ERROR,"com.read","Unable to initialise communications");
    return;
  }

  /* Wait for calculator */

  do {
    rdchar=COMsysGetChar();
  } while (rdchar!=CASPORT_TX_READY && rdchar!=CASPORT_TX_INLINE);
  COMsysPutChar(CASPORT_RX_READY);

  Msg(M_DEBUG,"com","Handshake successful");

  /* determine subformat */

  flags.input.subf=SUBF_UNKNOWN;

  if (args_is_set(dargs,"7700","9700","9800",NULL))
    flags.input.subf=SUBF_9700;
  if (args_is_set(dargs,"9750","9850","9950",NULL))
    flags.input.subf=SUBF_9850;
  if (args_is_set(dargs,"raw",NULL))
    flags.input.subf=SUBF_RAW;

  read_datalen=0;
  if (flags.input.subf==SUBF_UNKNOWN) {
    flags.input.subf=COMReadFirst(&hdr);
  } else {
    switch(flags.input.subf) {
      case SUBF_9850:
        hdr=tsys_malloc(HDR_SIZE_9850);
        COMReadBlock(hdr,HDR_SIZE_9850);
        break;
      case SUBF_9700:
        hdr=tsys_malloc(HDR_SIZE_9700);
        COMReadBlock(hdr,HDR_SIZE_9700);
        break;
      case SUBF_RAW: ;
      default: ;
    }
  }

  /* Receive data */

  switch (flags.input.subf) {
    case SUBF_9700: ReadData_9700(hdr, info, &io);break;
    case SUBF_9850: ReadData_9850(hdr, info, &io);break;
    case SUBF_RAW:  ReadData_Raw (hdr, info, &io);break;
  }

  COMsysClose();
}


void COMWriteData(char *filename,casinfo *info)
{
  darg_t dargs;
  int baud,parity,hand;
  int success;
  unsigned char rdchar;

  args_join(flags.output.args,ARGS_DEF_OUT,fCOM,&dargs);
  COMParseArgs(dargs, &baud, &parity, &hand);

  flg_bytes=args_is_set(dargs,"bytes",NULL);

  Msg(M_VERBOSE,"com","Settings: %s, %u, 8, %c, 2%s%s",filename,baud,parity,
    hand&CASPORT_DTR?", DTR":"",hand&CASPORT_RTS?", RTS":"");

  /* default CAS subfile type.  Should be more sophisticated, eg unsupported
  types should be raw etc.  Also output should be same as input by default */

  flags.output.subf=SUBF_9700;
  if (flags.input.format==fCAS||flags.input.format==fCOM)
    flags.output.subf=flags.input.subf;
  if (info->type==tUnknown||info->type==tRaw)
    flags.output.subf=SUBF_RAW;

  if (args_is_set(dargs,"9750","9850","9950",NULL))
    flags.output.subf=SUBF_9850;
  if (args_is_set(dargs,"raw",NULL)) flags.output.subf=SUBF_RAW;

  if (args_is_set(dargs,"pause",NULL)) {
    Msg(M_MSG,"com","press <enter> to begin communications");
    Pager(PAGER_FORCE|PAGER_NOW|PAGER_NOMSG);
  }

  /* Initialise COM port */

  if (!(COMsysInit(filename,baud,parity,hand))) {
    Msg(M_ERROR,"com.write","Unable to initialise communications");
    return;
  }

  /* Handshake with calculator */

  do {
    if (args_is_set(dargs,"inline",NULL)) {
      COMsysPutChar(CASPORT_TX_INLINE);
    } else {
      COMsysPutChar(CASPORT_TX_READY);
    }
    rdchar=COMsysGetCharTimeout(&success,100);

    if (rdchar==CASPORT_WRONGTYPE) {
      Msg(M_ERROR,"com","calculator refused handshake");
      COMsysClose();
      return;
    }
  } while (rdchar != CASPORT_RX_READY && rdchar != CASPORT_TX_INLINE);

  Msg(M_DEBUG,"com","Handshake successful");

  /* handle Send() data, where calc responds with type of data wanted */

  if (rdchar == CASPORT_TX_INLINE) {
    unsigned char recvdata[HDR_SIZE_9850];

    COMsysPutChar(CASPORT_RX_READY);
    COMReadBlock(recvdata,HDR_SIZE_9850);

    /* should act on this properly */
    Msg(M_DEBUG,"com.inline","calc requested data of type: `%c%c'",
            recvdata[4],recvdata[5]);
    rdchar=COMsysGetCharTimeout(&success,100);
    if (!success) {
      Msg(M_ERROR,"com.inline","calculator did not respond");
      COMsysClose();
      return;
    }
    if (rdchar!=CASPORT_OK) {
      Msg(M_ERROR,"com.inline","calculator cancelled transmission with `%02x'"
            ,rdchar);
      COMsysClose();
      return;
    } 
  }

  /* send data */

  switch (flags.output.subf) {
    case SUBF_9700: WriteData_9700(info, &io);break;
    case SUBF_9850: WriteData_9850(info, &io);break;
    case SUBF_RAW:  WriteData_Raw (info, &io);break;
  }

  COMsysClose();
}


void COMParseArgs(darg_t dargs, int *baud, int *parity, int *hand)
{

  /* defaults */
  (*parity)=CASPORT_EVEN;
  (*baud)=9600;
  (*hand)=CASPORT_DTR;

  if (args_is_set_keyed(dargs,"parity","even","E",NULL))
    (*parity)=CASPORT_EVEN;
  if (args_is_set_keyed(dargs,"parity","odd","O",NULL))
    (*parity)=CASPORT_ODD;
  if (args_is_set_keyed(dargs,"parity","none","N",NULL))
    (*parity)=CASPORT_NONE;

  if (args_is_set_keyed(dargs,"baud","9600",NULL))
    (*baud)=9600;
  if (args_is_set_keyed(dargs,"baud","4800",NULL))
    (*baud)=4800;
  if (args_is_set_keyed(dargs,"baud","2400",NULL))
    (*baud)=2400;
  if (args_is_set_keyed(dargs,"baud","1200",NULL))
    (*baud)=1200;

  if (args_is_set(dargs,"RTS"))
    (*hand) |= CASPORT_RTS;
  if (args_is_set(dargs,"DTR"))
    (*hand) |= CASPORT_DTR;
}

int COMReadFirst(void **hdr)
{
  darg_t dargs;
  unsigned char csum;
  unsigned char rdchar;
  unsigned char *data=NULL;
  long count;
  int success;

  args_join(flags.input.args,ARGS_DEF_IN,fCOM,&dargs);
  data=tsys_malloc(CAS_MEMSIZE);
  (*hdr)=data;

  /* Unlike fcas.c, which will read even a plain binary file, fcom.c
   * insists on the colon and csum mechanism.  Checksum failure only
   * results in a warning, however, not abort.
   */

  do {
    if ((rdchar=COMsysGetChar())!=':')
      Msg(M_WARN,"com/read","Expected `:', read 0x%02x",rdchar);
  } while (rdchar!=':');
  read_colon=':';

  /* Read in the first block in raw mode, as we don't know its length */

  count=0;
  for (;;) {
    rdchar=COMsysGetCharTimeout(&success,100);
    if (!success)
      break;

    /* Should have an expanding memory block, as in i=bin */
    if (count > CAS_MEMSIZE) {
      Msg(M_ERROR,"com.read","First block too large (%ld)",count);
      return SUBF_RAW;
    }
    data[count++]=rdchar;
  }
  read_csum=data[--count];      /* last byte is the checksum, not counted in length */
  read_datalen=count;

  if (args_is_set(dargs,"status",NULL)) {
    Msg(M_MSG,"com.readfirst","received block of size %ld (+2) bytes",read_datalen);
  }    

  COMsysPutChar(CASPORT_OK); /* maybe */

  csum=CalcCSum(data,count);
  if (read_csum!=csum)
    Msg(M_WARN,"com.csum","Checksum error.  expected 0x%02x, read 0x%02x",
                csum,read_csum);

  if (Is_9850(data)) {                           /* 9850 */
    Msg(M_DEBUG,"com.read","Autodetected subformat to be 9850");
    if (count != HDR_SIZE_9850)
      Msg(M_WARN,"com.read","Autodetected as 9850, yet header size mismatch (%d)",count);
    return SUBF_9850;

  } else if (Is_9700(data)) {
    Msg(M_DEBUG,"com.read","Autodetected subformat to be 9700");
    if (count != HDR_SIZE_9700)
      Msg(M_WARN,"com.read","Autodetected as 9700, yet header size mismatch (%d)",count);
    return SUBF_9700;

  } else {
    Msg(M_DEBUG,"com.read","Autodetected subformat to be raw");
    return SUBF_RAW;
  }
}

static int COMReadBlock(void *data,size_t size)
{
  darg_t dargs;
  unsigned char csum;
  unsigned char rdchar;
  unsigned char *curpos=NULL;
  long count;
  int flg_status=0;

  args_join(flags.input.args,ARGS_DEF_IN,fCOM,&dargs);

  flg_status=args_is_set(dargs,"status",NULL);

  do {
    if ((rdchar=COMsysGetChar())!=':')
      Msg(M_WARN,"com/read","Expected `:', read 0x%02x",rdchar);
  } while (rdchar!=':');
  read_colon=':';

  curpos=data;
  for (count=0;count<size;count++) {
    rdchar=COMsysGetChar();
    *curpos++=rdchar;
  };

  read_csum=COMsysGetChar();
  csum=CalcCSum(data,size);
  if (read_csum!=csum)
    Msg(M_WARN,"cas/csum","Checksum error.  expected 0x%02x, read 0x%02x",
                csum,read_csum);

  COMsysPutChar(CASPORT_OK);

  if (flg_status)
    Msg(M_MSG,"com.read","received block of size %ld (+2) bytes",size);

  return 1;
}


static int COMWriteBlock(void *data, size_t size)
{
  darg_t dargs;
  unsigned char *curpos=NULL;
  unsigned char ccsum;
  unsigned char rdchar;
  long count;
  int flg_status=0;
  int success;

  args_join(flags.output.args,ARGS_DEF_OUT,fCOM,&dargs);

  flg_status=args_is_set(dargs,"status",NULL);
  curpos=data;

  COMsysPutChar(':');
  for(count=0;count<size;count++) {
    COMsysPutChar(*curpos++);
  }
  ccsum=CalcCSum(data,size);
  COMsysPutChar(ccsum);

  if (flg_status)
    Msg(M_MSG,"com.write","transmitted block of %ld (+2) bytes",size);

  rdchar=COMsysGetChar();
#if 0
  rdchar=COMsysGetCharTimeout(&success,100);

  if (!success) {
    Msg(M_WARN,"com.write","calculator did not respond");
    return 0;
  }
#endif  
  return COMCheckResponse(rdchar);
}


static size_t COMReadRaw(void *dataptr, int trycook)
{
  darg_t dargs;
  unsigned int rdchar;
  int success=TRUE;
  long bsize=0;
  unsigned char *data;
  int flg_status=0;
  unsigned char csum;

  args_join(flags.input.args,ARGS_DEF_IN,fCOM,&dargs);
  flg_status=args_is_set(dargs,"status",NULL);
  data=dataptr;

  read_colon='\0';
  for (;;) {
    if (bsize>CAS_MEMSIZE) {
      Msg(M_WARN,"com.read","File too large, truncating (%ld bytes)",bsize);
      return bsize;
    }

    rdchar=COMsysGetCharTimeout(&success,100);

    if (!success) {     /* end of block */
      COMsysPutChar(CASPORT_OK);
      if (read_colon==':') {        /* cooked data, process checksum */
        read_csum=data[--bsize];
        csum=CalcCSum(data,bsize);
        if (read_csum!=csum) {
          Msg(M_WARN,"cas.readraw","Checksum error.  expected 0x%02x, read 0x%02x",
                  csum,read_csum);
        } else {
          /* If on an END block, and is header-sized +1 and the last char is
           * a 0, then pretty safe to assume it is the spurious last char
           * sometimes added by the calc at the end of transmission.  even
           * this pretty tight set of rules is probably not watertight
           */
          if (csum==0 && bsize==HDR_SIZE_9850 + 1 &&
                !strncmp((char *)data,"END",3) && data[3]==0xff) {
            read_csum=data[--bsize];
            Msg(M_DEBUG,"cas.readraw","removed strongly suspected spurious 00 at end of block");
          }
        }
      }
      if (flg_status) {
        if (read_colon==':')
          Msg(M_MSG,"com.readraw","received block of size %ld (+2) bytes",bsize);
        else
          Msg(M_MSG,"com.readraw","received block of size %ld (+0) bytes",bsize);
      }
      return bsize;

    } else {

      if (bsize==0 && trycook) {
        read_colon=rdchar;

        /* if first char not a colon, then assume non-cooked data */
        if (read_colon != ':') data[bsize++]=rdchar;

        if (read_colon==':')
          Msg(M_DEBUG,"com.readraw","Assuming cooked data (first char `:')");
        else
          Msg(M_DEBUG,"com.readraw","Assuming raw data (first char '0x%02x')",read_colon);

        /* make sure we don't test char #2 for cookedness as well... */
        trycook=0;
      } else {
        data[bsize++]=rdchar;
      }
    }
  }
}

static int COMWriteRaw(void *data, size_t datalen)
{
  darg_t dargs;
  unsigned int rdchar;
  long bsize,i;
  int success;
  unsigned char *curpos;
  int flg_status;

  args_join(flags.output.args,ARGS_DEF_OUT,fCOM,&dargs);

  flg_status=args_is_set(dargs,"status",NULL);
  curpos=data;

  bsize=0;
  for (i=0;i<datalen;i++) {
    COMsysPutChar(curpos[i]);
    bsize++;
    /* New blocks always start with a colon, however, not *every* colon
     * is the start of a new block, so we must have the timeout.
     */
    if (i+1==datalen || curpos[i+1]==':') {
      rdchar=COMsysGetCharTimeout(&success,100);
      if (success) {
        if (flg_status)
          Msg(M_MSG,"com.write","transmitted block of size %ld (+2) bytes",bsize);
        bsize=0;
      }
    }
  }
  Msg(M_VERBOSE,"com.writeraw","Transmitted data");

  return 1;
}

/* returns 0 if unable to continue */

int COMCheckResponse(unsigned char response)
{
  darg_t dargs;
  unsigned char rdchar='\xff';

  args_join(flags.output.args,ARGS_DEF_OUT,fCOM,&dargs);

  switch (response) {
    case CASPORT_OK_9850:
    case CASPORT_OK:
             return 1;
    case CASPORT_BADCSUM:
             Msg(M_ERROR,"com.response","Checksum error");
             return 0;
    case CASPORT_AREAFULL:
             if (args_is_set(dargs,"overwrite","over",NULL)) {
               COMsysPutChar(CASPORT_OVERWRITE);
               if ((rdchar=COMsysGetChar())!=CASPORT_OK) {
                 Msg(M_ERROR,"com.response","Unable to overwrite program: 0x%02x",rdchar);
                 return 0;
               }
               Msg(M_DEBUG,"com.response","Overwriting existing program");
               return 1;
             } else {
               COMsysPutChar(CASPORT_CANCEL);
               Msg(M_ERROR,"com.response","Program area already in use");
               return 0;
             }
    case CASPORT_MEMFULL:
             Msg(M_ERROR,"com.response","Calculator is out of free memory");
             return 0;
    case CASPORT_WRONGTYPE:
             Msg(M_ERROR,"com.response","Wrong kind of data");
             return 0;
    case CASPORT_BADFILE:
             Msg(M_ERROR,"com.response","Corrupted CAS file");
             return 0;
    default: Msg(M_WARN,"com.response","Unknown response code: 0x%02x",response);
  }
  return 1;  /* carry on anyway */
}

#else /* CAS_COMM */

void COMReadData(char *filename, casinfo *info_r)
{
  Msg(M_ERROR,"com.read","Serial communications not yet supported on this platform");
}

void COMWriteData(char *filename,casinfo *info)
{
  Msg(M_ERROR,"com.write","Serial communications not yet supported on this platform");
}

#endif /* CAS_COMM */

/* end of fcom.c */
