/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * MYLIBC.C  Miscellaneous general purpose utilities + std library wrappers
 */

#define _MYLIB_C_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mylibc.h"

/*
 *  Memory allocation routines with error checking, reporting and exit
 */

void *tsys_malloc (size_t size)
{
  void *memptr=NULL;
  
  if ((memptr=malloc(size)) == NULL) {
    fprintf(stderr,"malloc: unable to allocate %ld bytes",(long)size);
    exit(EXIT_MEM);
  }

  return memptr;
}
         
void *tsys_calloc (size_t size)
{
  void *memptr=NULL;
  
  if ((memptr=calloc(1,size)) == NULL) {
    fprintf(stderr,"calloc: unable to allocate %ld bytes",(long)size);
    exit(EXIT_MEM);
  }

  return memptr;
}              

void *tsys_realloc (void *oldptr, size_t size)
{
  void *memptr=NULL;
  
  if (size==0) {
    free(oldptr);
    return NULL;
  } else {
    if ((memptr=realloc(oldptr, size)) == NULL) {
      fprintf(stderr,"realloc: unable to allocate %ld bytes",(long)size);
      exit(EXIT_MEM);
    }
  }

  return memptr;
}

FILE *tsys_fopen(char *fname, char *faccess, int owndir)
{
  char newname[2 * FILENAME_MAX+1];		/* large enough for worst case */
  char *eod=NULL;
  FILE *handle=NULL;

  if (!fname||!faccess) return NULL;

  if (!strcmp(fname,"stdin")) return stdin;
  if (!strcmp(fname,"stdout")) return stdout;
  if (!strcmp(fname,"stderr")) return stderr;
  if (!strcmp(fname,"-")) {
    switch(*faccess) {
      case 'w': return stdout;
      case 'r': return stdin;
      default: return NULL;
    }
  }

  if ((handle=fopen(fname,faccess))) return handle;
  if (!owndir) return NULL;

#ifdef CAS_DOS
  strcpy(newname,argv0);
  if (!(eod=strrchr(newname,CAS_DIRSEP)))
    return NULL;    
  strcpy(eod+1,fname);
#else
# ifdef CAS_UNIX
  strcpy(newname,CAS_UNIX_HOMEDIR);
  if (!(eod=strrchr(newname,CAS_DIRSEP)))
    return NULL;    
  strcpy(eod+1,fname);
# else
  return NULL;
# endif
#endif

  if ((handle=fopen(newname,faccess))) return handle;
  return NULL;
}

char *tsys_strdup(char *s1)
{
  char *s2;

  if (!s1) return NULL;
  s2 = tsys_malloc(strlen(s1) + 1);
  strcpy(s2, s1);

  return s2;
}

char *strdup_trim(char *s1, char *strip)
{
  char *s2;
  char *i;

  /* strip leading characters by incrementing pointer to original string */
  for (i=s1;*i != 0; ++i) {
    if (strchr(strip,*i) != NULL)
      s1++;
    else
      break;
  }

  s2=tsys_strdup(s1);

  /* convert all trailing strippable characters in the copy to '\0' */
  for (i=s2+strlen(s2)-1; i > s2; ++i) {
    if (strchr(strip,*i) != NULL)
      *i='\0';
    else
      break;
  }

  return s2;
}

/* NULL safe strcmp, stricmp, strncmp, strnicmp */

int tsys_strcmp(const char *s1, const char * s2)
{
  if (s1 == NULL && s2 == NULL) return 0;
  if (s1 == NULL) return -1;
  if (s2 == NULL) return 1;
  
  return strcmp(s1, s2);
}

int tsys_strncmp(const char *s1, const char *s2, int n)
{
  if (s1 == NULL && s2 == NULL) return 0;
  if (s1 == NULL) return -1;
  if (s2 == NULL) return 1;
  
  return strncmp(s1, s2, n);
}

int tsys_stricmp(const char *s1, const char *s2)
{
  if (s1 == NULL && s2 == NULL) return 0;
  if (s1 == NULL) return -1;
  if (s2 == NULL) return 1;
  
  return stricmp(s1, s2);
}

int tsys_strnicmp(const char *s1, const char *s2, int n)
{
  if (s1 == NULL && s2 == NULL) return 0;
  if (s1 == NULL) return -1;
  if (s2 == NULL) return 1;
  
  return strnicmp(s1, s2, n);
}

int tsys_fclose(FILE *handle)
{
  if (!handle) return EOF;
  if (handle==stdin) return 0;
  if (handle==stdout) return 0;
  if (handle==stderr) return 0;

  return fclose(handle);
}


/*
 * Calculate checksums
 */

int csum_ignore_first=0;

unsigned char CalcCSum(void *data, int size)
{
  register unsigned char cs=0;

  if (!data || size<=0) return 0;

  if (csum_ignore_first) {
    ((unsigned char *)data)++;size--;
  }

  csum_ignore_first=0;

  for (;size;)
    cs+=256-((unsigned char*)data)[--size];

  return cs & 0xff;
}

/*
 *  recursive wildcard matching function.  Only works with * and ? so far :(
 */

int WildMatch(char *wildcard, char *teststr)
{
  char *nextw=NULL;
  char *nextt=NULL;
  
  if (!*teststr && !*wildcard) return TRUE;
  if (!*teststr || !*wildcard) return FALSE;
  
  if (*wildcard=='?'||(*teststr==*wildcard)) {
    return(WildMatch(++wildcard,++teststr));
  }
  if (*wildcard=='*') {
    nextw=wildcard;
    if (!*++nextw) return TRUE;
    if (!(nextt=strchr(teststr,*nextw))) return FALSE;
    if (WildMatch(nextw,nextt)) return TRUE;
    return (WildMatch(wildcard,++nextt));
  }
  return FALSE;
}


/* The ELF hash algorithm */

unsigned long Hash(char *s, unsigned max)
{
  register unsigned long h = 0;
  register unsigned long g;

  while (*s) {
    h = (h << 4) + *s++;
    g = h & 0xf0000000L;
    if (g) h ^= g >> 24;
    h &= ~g;
  }

  return h % max;
}

/* another hashing algorithm, used for probing */

unsigned long Hash2(char *s, unsigned max)
{
 register unsigned long hashval;
 
 for (hashval=0;*s;s++)
   hashval = *s * 7 + hashval * 15;
 return (hashval%max);
}



/*
 * Convert numbers depending on endianness.  Stolen from QEU 0.3.
 */
#ifdef FAT_ENDIAN

ulongint SwapInt32(ulongint x)
{
  return (((x << 24) & 0xff000000)
          | ((x<< 8) & 0x00ff0000)
          | ((x>> 8) & 0x0000ff00)
          | ((x>>24) & 0x000000ff));
}

#endif /* FAT_ENDIAN */

/* end of mylibc.c */

