/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * S9700.C  Routines for translating the 7700, 9700, 9800 .CAS format
 */
 
#define _CAS_S9700_C_

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "list.h"
#include "s9700.h"


void ReadData_9700(void *firsthdr, casinfo *info, iostruct *io)
{
  casheader *hdr;
  int match;

  hdr=firsthdr;
  
  for (match=0;match<sizeof castypes/sizeof(castypes_t);match++) {
    if (!strncmp(castypes[match].magic,(char *)hdr->magic,2)) {
      (*castypes[match].read)(hdr,info,io);
      goto done;
    }
  }

  ReadRaw(hdr,info,io);

  done:
  free(hdr);
}


void WriteData_9700(casinfo *info, iostruct *io)
{
  for(;info;info=info->next) {
    switch(info->type) {
      case tOldProg:
        if (info->s.prg.prognum>=0 && info->s.prg.prognum<CAS_PZ_NUMPROGS)
          WritePZ(info,io);
        else
          WriteP1(info,io);
        return;

      case tEditor:
        if (IsMultiple(info))
          WriteFN(info,io);
        else
          WriteEN(info,io);
        return;

      case tSSMono:
        WriteDD(info,io);
        return;

      case tSSCol:
        WriteDC(info,io);
        return;

      case tRaw:
        WriteRaw(info,io);
        return;

      case tDesc:
      case tText:
        break;
      case tEnd:
      case tUnknown:
        Msg(M_WARN,"9700.write","strange unwritable block (type: %s)",
                                            s_types[info->type]);
        break;

      default:
        Msg(M_ERROR,"9700.write","unable to write block of type: %s",
                                            s_types[info->type]);
    }
    /* loop round and see if the next block is writable */
  }
}


static int IsMultiple(casinfo *info)
{
  types curtype;

  curtype=info->type;
  for(info=info->next;info;info=info->next) {
    if (info->type==curtype) return TRUE;
  }

  return FALSE;
}

int Is_9700(unsigned char *data)
{
  int match;

  for (match=0;match<sizeof castypes/sizeof(castypes_t);match++) {
    if (!strncmp((char *)data,castypes[match].magic,2)) {
      return 1;
    }
  }
  /* no match */
  return 0;
}



/*
 * Functions to instantiate the basic data structures
 */

static casinfo *PrepareInfo(types type, size_t datalen)
{
  casinfo *info;

  info=tsys_malloc(sizeof(casinfo));

  info->type=type;
  info->length=0;       /* should only be set when data *present*  */
  info->next=NULL;
  if (datalen)
    info->data=tsys_malloc(datalen);
  else
    info->data=NULL;

  /* specific headers should be set by calling routine */

  return info;
}


static casheader *PrepareHeader(char magic[2], char misc)
{
  int i;
  casheader *hdr;

  hdr=tsys_malloc(sizeof(casheader));

  for (i=0;i< 2;i++) hdr->magic[i]=magic[i];
  for (i=0;i< 5;i++) hdr->misc[i]=misc;
  for (i=0;i<12;i++) hdr->filename[i]='\xff';
  for (i=0;i<12;i++) hdr->password[i]='\xff';
  for (i=0;i< 7;i++) hdr->reserved[i]='\xff';

  return hdr;
}


/*
 * Raw data
 */


static void ReadRaw(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  unsigned char *data;
  size_t datalen;

  /* copy first header into a raw block */
  curblk=info->next=PrepareInfo(tRaw,sizeof (casheader));
  curblk->s.raw.cooked=1;
  memcpy(curblk->data,hdr,sizeof (casheader));
  curblk->length=sizeof (casheader);

  /* copy all following blocks into raw blocks */

  data=tsys_malloc(CAS_MEMSIZE);
  for (;;) {
    datalen=(*io->readraw)(data, 1);

    /* screen out those annoying length=1 blocks often at end of trans */
    if (datalen<=1) {
      break;
    }

    curblk=curblk->next=PrepareInfo(tRaw,datalen);
    if (read_colon == ':') {                      /* cooked data */
      curblk->s.raw.cooked=1;
    } else {
      curblk->s.raw.cooked=0;
    }

    memcpy(curblk->data,data,datalen);
    curblk->length=datalen;
    curblk->next=NULL;
  }
  free(data);
}

static void WriteRaw(casinfo *info, iostruct *io)
{
  /* this will only output all the raw blocks then exit */

  for(;info;info=info->next) {
    if (info->type==tRaw && info->length > 0) {
      if (info->type == tRaw && info->s.raw.cooked)
        (*io->writeblk)(info->data,info->length);
      else
        (*io->writeraw)(info->data,info->length);
    }
  }
  return;
}


/*
 *  Single Program
 */


static void ReadP1(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  unsigned char *data;
  size_t datalen;

  datalen=(hdr->misc[1] << 8) + hdr->misc[2] - 2;
  curblk=info->next=PrepareInfo(tOldProg,datalen);
  data=curblk->data;
  if ((*io->readblk)(data,datalen) == 0) {
    Msg(M_WARN,"cas.9700.read","Unable to read single program block..skipping");
    info->next=NULL;
    free(curblk->data);
    free(curblk);
    return;
  }
  curblk->length=datalen;
  curblk->s.prg.prognum=-1;
  curblk->s.prg.progtype=hdr->misc[3];

  free(hdr);
}

static void WriteP1(casinfo *info, iostruct *io)
{
  casheader *hdr;

  hdr=PrepareHeader("P1",'\0');

  hdr->misc[1]=((info->length+2) & 0xff00) >> 8;
  hdr->misc[2]=((info->length+2) & 0x00ff);
  hdr->misc[3]=info->s.prg.progtype;
  if ((*io->writeblk)(hdr,sizeof(casheader)) == 0) {
    Msg(M_WARN,"cas.9700.write","Unable to write header..skipping block");
    return;
  }
  if ((*io->writeblk)(info->data,info->length) == 0) {
    Msg(M_WARN,"cas.9700.write","Unable to write single program block..skipping");
  }
  return;
}


/*
 * Multiple programs
 */


static void ReadPZ(casheader *hdr, casinfo *info, iostruct *io)
{
  prgheader ph[CAS_PZ_NUMPROGS];
  int datalen;
  int blklen;
  int n;
  unsigned char *buffer;
  unsigned char *data;
  casinfo *curblk;

  if ((*io->readblk)(ph,sizeof ph) == 0) {
    Msg(M_WARN,"9700.read","Unable to read multiple program info header..skipping");
    return;
  }

  datalen = (hdr->misc[1] << 8) + hdr->misc[2];
  buffer=tsys_malloc(datalen - 2);

  if ((*io->readblk)(buffer,datalen - 2) == 0) {
    Msg(M_WARN,"9700.read","Unable to read multiple program data..skipping");
    free(buffer);
    return;
  }

  data=buffer;
  curblk=info;

  for (n=0;n<CAS_PZ_NUMPROGS;n++) {
    blklen=(ph[n].sizehi << 8) + ph[n].sizelo;
    if (blklen>1) {
      curblk=curblk->next=PrepareInfo(tOldProg,blklen);
      memcpy(curblk->data,data,blklen);
      curblk->length=blklen;
      curblk->s.prg.prognum=n;
      curblk->s.prg.progtype=ph[n].progtype;
    }
    data+=blklen;
  }

  curblk->next=NULL; /* security */
  free(buffer);
}


static void WritePZ(casinfo *info, iostruct *io)
{
  int n;
  prgheader ph[CAS_PZ_NUMPROGS];
  unsigned char *dataptrs[CAS_PZ_NUMPROGS];
  unsigned char *buffer;
  unsigned char *curpos;
  casheader *hdr;
  long datalen;

  hdr=PrepareHeader("PZ",'\0');

  /* initialise header and array of data pointers */

  for (n=0;n<CAS_PZ_NUMPROGS;n++) {
    ph[n].sizehi=0;
    ph[n].sizelo=1;
    ph[n].progtype=0;
    ph[n].unknown1=0;
    ph[n].unknown3=0;
    dataptrs[n]=NULL;
  }

  datalen=CAS_PZ_NUMPROGS;

  /* load info about current set of programs into header and pointers */

  for(;info;info=info->next) {
    switch (info->type) {
      case tOldProg:
        if (info->s.prg.prognum<0 || info->s.prg.prognum>CAS_PZ_NUMPROGS) {
          Msg(M_WARN,"9700.write","Program has no identifier..skipping");
          break;
        }
        if (info->length <= 1) {
          Msg(M_WARN,"9700.write","No data in program #%d..skipping",
                                            info->s.prg.prognum);
          break;
        }
        datalen += info->length - 1;
        ph[info->s.prg.prognum].sizehi=(info->length >> 8) & 0xff;
        ph[info->s.prg.prognum].sizelo=(info->length     ) & 0xff;
        ph[info->s.prg.prognum].progtype=info->s.prg.progtype;
        dataptrs[info->s.prg.prognum]=info->data;
        break;

      case tText:
      case tDesc:
        break;
      default:
        Msg(M_WARN,"9700.write","cannot include data of type %s in multiple program",
            s_types[info->type]);
    }
  }

  /* copy data into single buffer in memory */

  buffer=tsys_malloc(datalen);
  hdr->misc[1]=((datalen + 2) >> 8) & 255;
  hdr->misc[2]=((datalen + 2)     ) & 255;
  curpos=buffer;

  for (n=0;n<CAS_PZ_NUMPROGS;n++) {
    if (dataptrs[n] != NULL) {
      long blklen=((long)(ph[n].sizehi) << 8) + ph[n].sizelo;

      memcpy(curpos,dataptrs[n],blklen);
      curpos+=blklen;
    } else {
      *curpos++=0xff;
    }
  }

  /* write data */

  (*io->writeblk)(hdr,sizeof (casheader));
  (*io->writeblk)(ph,sizeof ph);
  (*io->writeblk)(buffer,datalen);

  free(hdr);
  free(buffer);
}


/*
 *  Single Editor files
 */


static void ReadEN(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  long datalen;

  datalen=(hdr->misc[1] << 8) + hdr->misc[2] - 2;

  curblk=PrepareInfo(tEditor,datalen);

  curblk->s.ed.filename[12]='\0';
  memcpy(curblk->s.ed.filename,hdr->filename,12);

  curblk->s.ed.password[12]='\0';
  memcpy(curblk->s.ed.password,hdr->password,12);

  if ((*io->readblk)(curblk->data,datalen) == 0) {
    Msg(M_WARN,"9700.read","Unable to read editor data block..skipping");
    free(curblk->data);
    free(curblk);
    return;
  }
  curblk->length=datalen;
  info->next=curblk;
}

static void WriteEN(casinfo *info, iostruct *io)
{
  casheader *hdr;

  hdr=PrepareHeader("EN",'\0');

  hdr->misc[1]=((info->length + 2) & 0xff00) >> 8;
  hdr->misc[2]=((info->length + 2) & 0x00ff);

  memcpy(hdr->filename,info->s.ed.filename,12);
  memcpy(hdr->password,info->s.ed.password,12);

  (*io->writeblk)(hdr,sizeof(casheader));
  (*io->writeblk)(info->data,info->length);

  free(hdr);
}


/*
 *  Multiple Editor Files
 */


static void ReadFN(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  casinfo *prevblk;
  int blklen;

  curblk=info;

  while (hdr->magic[0]=='F' && (hdr->magic[1]=='N' || hdr->magic[1]=='P')) {
    blklen=((hdr->misc[1]) << 8) +  hdr->misc[2] - 2;
    prevblk=curblk;
    curblk=prevblk->next=PrepareInfo(tEditor,blklen);
    curblk->s.ed.filename[12]='\0';
    memcpy(curblk->s.ed.filename,hdr->filename,12);
    curblk->s.ed.password[12]='\0';
    memcpy(curblk->s.ed.password,hdr->password,12);

    if ((*io->readblk)(curblk->data,blklen)==0) {
      Msg(M_WARN,"9700.read","Unable to read editor data block..skipping");
      prevblk->next=NULL;
      free(curblk->data);
      free(curblk);
      free(hdr);
      return;
    }

    curblk->length=blklen;

    if ((*io->readblk)(hdr,sizeof(casheader))==0) {
      Msg(M_WARN,"9700.read","Unable to read header for next block..skipping");
      return;
    }
  }
}

static void WriteFN(casinfo *info, iostruct *io)
{
  casinfo *curblk;
  casheader *hdr;

  hdr=PrepareHeader("FN",'\0');

  for(curblk=info;curblk;curblk=curblk->next) {
    switch(curblk->type) {
      case tEditor:
        hdr->misc[1]=((curblk->length + 2) & 0xff00) >> 8;
        hdr->misc[2]=((curblk->length + 2) & 0x00ff);
        
        memcpy(hdr->filename,curblk->s.ed.filename,12);
        memcpy(hdr->password,curblk->s.ed.password,12);

        if ((*io->writeblk)(hdr,sizeof(casheader)) == 0) {
          Msg(M_WARN,"9700.write","Unable to write header for editor file..skipping");
          break;
        }
        if ((*io->writeblk)(curblk->data,curblk->length) ==0) {
          Msg(M_WARN,"9700.write","unable to write editor file..aborting");
          free(hdr);
          return;
        }
        break;
      case tDesc:
      case tText:
        break;
      default:
        Msg(M_WARN,"9700.write","Cannot include data of type %s in multiple editor",
            s_types[info->type]);
    }
  }
  free(hdr);
  hdr=PrepareHeader("\x17\xff",'\xff');
  if ((*io->writeblk)(hdr,sizeof(casheader)) == 0) {
    Msg(M_WARN,"9700.write","unable to write end block");
  }
  free(hdr);
}


/*
 *  Screenshot helper functions
 */

extern int csum_ignore_first;

static int read_ss_block(casinfo *info, iostruct *io)
{
  darg_t dargs;
  unsigned char *readblk;
  unsigned char *datablk;
  unsigned char *data;
  unsigned char colour;
  long offset;
  int flg_inv;
  int x,y;
  div_t x8;
  int nbw;          /* number of blocks wide */
  long datalen;

  args_join(flags.input.args,ARGS_DEF_IN,fCAS,&dargs);

  /* read data */

  nbw=((info->s.ss.width + 7) & ~7) / 8;
  datalen=nbw * info->s.ss.height;

  Msg(M_DEBUG,"9700.read.ss","width=%d, height=%d, datalen=%ld, nbw=%d",
                info->s.ss.width, info->s.ss.height, datalen, nbw);

  if (info->s.ss.bpp<=1) {
    readblk=tsys_malloc(datalen);
    if ((*io->readblk)(readblk,datalen) == 0) {
      Msg(M_ERROR,"9700.read","Unable to read screenshot data");
      free(readblk);
      return 0;
    }
    datablk=readblk;
    colour=1;
  } else {
    csum_ignore_first=1;
    readblk=tsys_malloc(datalen + 1);
    if ((*io->readblk)(readblk,datalen + 1)==0) {
      Msg(M_ERROR,"9700.read","Unable to read colour screenshot data");
      free(readblk);
      return 0;
    }
    datablk=readblk+1;
    colour=*readblk;
  }

  flg_inv=args_is_set(dargs,"inverse","inv",NULL);
  data=info->data;

  for (y=0;y<info->s.ss.height;y++) {
    for (x=0;x<info->s.ss.width;x++) {
      x8=div(x,8);
      switch(info->s.ss.order) {
        case SS_ORDER_9700:
          offset=(nbw - x8.quot - 1) * info->s.ss.height + y;
          break;
        case SS_ORDER_9800:
          offset=x8.quot + y * nbw;
          break;
        default:
          Msg(M_ERROR,"9700.read","Unknown screenshot data format '%c'",
            info->s.ss.order);
          free(readblk);
          return 0;
      }
      if (datablk[offset] & pow2[7-x8.rem]) {
        data[SSOFFSET(info,x,y)] += flg_inv ? 0 : colour;
      } else {
        data[SSOFFSET(info,x,y)] += flg_inv ? colour : 0;
      }
    }
  }

  free(readblk);
  return 1;
}

static int write_ss_block(casinfo *info, unsigned char col, iostruct *io)
{
  darg_t dargs;
  unsigned char *data;
  unsigned char *writeblk;
  unsigned char *datablk;
  long datalen;
  int nbw;
  div_t x8;
  int x,y;
  long offset;
  int flg_inv;

  args_join(flags.output.args,ARGS_DEF_OUT,fCAS,&dargs);

  nbw=((info->s.ss.width + 7) & ~7) / 8;
    /* We expand the screenshot to be a multiple of 8 pixels wide. The extra
     * pixels should be 0, but it doesn't really matter.
     */

  datalen=nbw * info->s.ss.height;
  writeblk=datablk=tsys_calloc(datalen+1);    /* initialised to 0 */

  if (col!=(unsigned char)0xff) {   /* col=='\xff' symbolises b+w */
    *datablk++=col & 255;
    ++datalen;
    csum_ignore_first=1;
  }

  flg_inv=args_is_set(dargs,"inverse","inv",NULL);
  data=info->data;

  for (y=0;y<info->s.ss.height;y++) {
    for (x=0;x<info->s.ss.width;x++) {
      x8=div(x,8);
      switch(info->s.ss.order) {
        case SS_ORDER_9700:
          offset=(nbw - x8.quot - 1) * info->s.ss.height + y;
          break;
        case SS_ORDER_9800:
          offset=x8.quot + y * nbw;
          break;
        default:
          Msg(M_ERROR,"9700.read","Unknown screenshot data format '%c'",
            info->s.ss.order);
          free(writeblk);
          return 0;
      }

      if ((data[SSOFFSET(info,x,y)] & col) != 0) {
        datablk[offset] |= flg_inv ? 0 : pow2[7-x8.rem];
      } else {
        datablk[offset] |= flg_inv ? pow2[7-x8.rem] : 0;
      }
    }
  }

  if ((*io->writeblk)(writeblk,datalen) == 0) {
    Msg(M_WARN,"9700.write","Unable to write block of screenshot");
    free(writeblk);
    return 0;
  }

  return 1;
}

/*
 *  Screenshots
 */

static void ReadDD(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  long datalen;

  datalen=hdr->misc[1] * hdr->misc[0];
  Msg(M_DEBUG,"9700.read","ss datalen=%ld",datalen);
  curblk=PrepareInfo(tSSMono,datalen);
  memset(curblk->data,0,datalen);

  curblk->s.ss.width=hdr->misc[1];
  curblk->s.ss.height=hdr->misc[0];
  curblk->s.ss.bpp=1;
  curblk->s.ss.order=hdr->misc[3];

  if (read_ss_block(curblk,io) == 0) {
    Msg(M_WARN,"9700.read","Unable to read screenshot..skipping");
    free(curblk->data);
    free(curblk);
  }

  curblk->length=datalen;
  info->next=curblk;
}


static void WriteDD(casinfo *info, iostruct *io)
{
  darg_t dargs;
  casheader *hdr;

  args_join(flags.output.args,ARGS_DEF_OUT,fCAS,&dargs);

  if (info->s.ss.order==SS_ORDER_UNKNOWN)
    info->s.ss.order=SS_ORDER_DEFAULT;

  if (!stricmp(args_get_string(dargs,"order",NULL),"9700"))
    info->s.ss.order=SS_ORDER_9700;
  if (!stricmp(args_get_string(dargs,"order",NULL),"9800"))
    info->s.ss.order=SS_ORDER_9800;

  hdr=PrepareHeader("DD",'\0');

  hdr->misc[0]=info->s.ss.height;
  hdr->misc[1]=info->s.ss.width;
  if (info->s.ss.order==SS_ORDER_9700)
    hdr->misc[2]=0x10;
  else
    hdr->misc[2]=0;            /* unknown */
  hdr->misc[3]=info->s.ss.order;
  hdr->misc[4]='W';            /* unknown */
  hdr->filename[0]='F';        /* unknown */

  if ((*io->writeblk)(hdr,sizeof(casheader))==0) {
    Msg(M_WARN,"9700.write","Unable to write screenshot header...skipping");
    free(hdr);
    return;
  }
  free(hdr);

  write_ss_block(info,'\xff',io);
}


static void ReadDC(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  long datalen;

  datalen=hdr->misc[1] * hdr->misc[0];
  curblk=PrepareInfo(tSSCol,datalen);
  memset(curblk->data,0,datalen);

  curblk->s.ss.width=hdr->misc[1];
  curblk->s.ss.height=hdr->misc[0];
  curblk->s.ss.bpp=3;
  curblk->s.ss.order=hdr->misc[3];

  if (read_ss_block(curblk,io) == 0) {
    Msg(M_WARN,"9700.read","Unable to read screenshot data 1..skipping");
    free(curblk->data);
    free(curblk);
  }
  if (read_ss_block(curblk,io) == 0) {
    Msg(M_WARN,"9700.read","Unable to read screenshot data 2..skipping");
    free(curblk->data);
    free(curblk);
  }
  if (read_ss_block(curblk,io) == 0) {
    Msg(M_WARN,"9700.read","Unable to read screenshot data 3..skipping");
    free(curblk->data);
    free(curblk);
  }

  info->next=curblk;
}


static void WriteDC(casinfo *info, iostruct *io)
{
  darg_t dargs;
  casheader *hdr;

  args_join(flags.output.args,ARGS_DEF_OUT,fCAS,&dargs);

  if (info->s.ss.order==SS_ORDER_UNKNOWN)
    info->s.ss.order=SS_ORDER_DEFAULT;

  if (!stricmp(args_get_string(dargs,"order",NULL),"9700"))
    info->s.ss.order=SS_ORDER_9700;
  if (!stricmp(args_get_string(dargs,"order",NULL),"9800"))
    info->s.ss.order=SS_ORDER_9800;

  hdr=PrepareHeader("DC",'\0');

  hdr->misc[0]=info->s.ss.height;
  hdr->misc[1]=info->s.ss.width;
  if (info->s.ss.order==SS_ORDER_9700)
    hdr->misc[2]=0x10;
  else
    hdr->misc[2]=0;            /* unknown */
  hdr->misc[3]=info->s.ss.order;
  hdr->misc[4]='W';            /* unknown */
  hdr->filename[0]='F';        /* unknown */
  hdr->filename[1]='\003';      /* unknown; number of blocks? */

  if ((*io->writeblk)(hdr,sizeof(casheader))==0) {
    Msg(M_WARN,"9700.write","Unable to write screenshot header...skipping");
    free(hdr);
    return;
  }
  free(hdr);

  if (write_ss_block(info,1,io)==0)
    Msg(M_WARN,"9700.write","Unable to write blue screenshot block");
  if (write_ss_block(info,2,io)==0)
    Msg(M_WARN,"9700.write","Unable to write green screenshot block");
  if (write_ss_block(info,4,io)==0)
    Msg(M_WARN,"9700.write","Unable to write red screenshot block");
}


/* end of s9700.c */
