/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * TOKEN.C  Routines for loading token files, and converting to and from
 *          text tokens and integer ctokens
 */

/*
 * define CAS_TOKCHECK for debugging info. (as in tokcheck)
 */

#define _CAS_TOKEN_C_

#include "cas.h"
#include "misc.h"
#include "token.h"

static int clash=0;

/*
 * InitTokens.  Allocates memory for the arrays of pointers to the tokens
 *              and calls LoadTokens appropriately            
 */

void token_init(char *in_tokf, char *out_tokf, char *list_tokf)
{
  int i;

  toklis = tsys_calloc(sizeof (tokenlist));
  Msg(M_DEBUG,"token","Allocated 0x%04lx bytes for table", sizeof (tokenlist)); 
  toklis->tokpos = toklis->tokblock = tsys_malloc(TOKEN_BLOCKSIZE);
  Msg(M_DEBUG,"token","Allocated 0x%04lx bytes for tokens", TOKEN_BLOCKSIZE); 
  toklis->tokend = toklis->tokblock + TOKEN_BLOCKSIZE;
  for (i=0; i<TOKEN_MAX; i++) {
    toklis->out_nice[i] = NULL;
    toklis->out_str[i] = NULL;
    toklis->list_nice[i] = NULL;
    toklis->list_str[i] = NULL;
  }

  if (in_tokf != NULL && *in_tokf != '\0') {
    token_load(in_tokf, 1, 1, 1);
  }

  if (out_tokf != NULL && *out_tokf != '\0' &&
      tsys_strcmp(in_tokf, out_tokf) != 0) {
    if (in_tokf != NULL && *in_tokf != '\0') {
      for (i=0; i<TOKEN_MAX; i++) {
        toklis->out_nice[i] = NULL;
        toklis->out_str[i] = NULL;
      }
    }
    if (tsys_strcmp(out_tokf, list_tokf) == 0) {
      if (in_tokf != NULL && *in_tokf != '\0') {
        for (i=0;i<TOKEN_MAX; i++) {
          toklis->list_nice[i] = NULL;
          toklis->list_str[i] = NULL;
        }
      }
      token_load(out_tokf, 0, 1, 1);
    } else {
      token_load(out_tokf, 0, 1, 0);
    }
  }

  if (list_tokf != NULL && *list_tokf != '\0'
                && tsys_strcmp(in_tokf, list_tokf) != 0
                && tsys_strcmp(out_tokf, list_tokf) != 0) {
    if ( (in_tokf != NULL && *in_tokf != '\0') || 
         (out_tokf != NULL && *out_tokf != '\0')) {
      for (i=0; i<TOKEN_MAX; i++) {
        toklis->list_nice[i] = NULL;
        toklis->list_str[i] = NULL;
      }
    }
    token_load(list_tokf, 0, 0, 1);
  }
}

/*
 * LoadTokens.  Loads all the tokens from a tokenfile into the respective
 *              structures determined by in, out and list.
 */

void token_load(char *tokfile,int in,int out,int list)
{
  FILE *tf=NULL;
  char tokline[TOKEN_LINESIZE];
  int count=0;
  int numtok=0;
  int linecount;

  tf = tsys_fopen(tokfile,"r",TRUE);
  if (!tf) {
    Msg(EXIT_FILE,"token","unable to read token file `%s'",tokfile);
  }
  
  for(;;) {
    if (feof(tf))
      break;
    fgets(tokline, TOKEN_LINESIZE, tf);
    if (tokline[strlen(tokline) - 1] == '\n')
      tokline[strlen(tokline) - 1] = '\0';
#ifdef CAS_TOKCHECK
    if (isspace(tokline[strlen(tokline) - 1]))
      Msg(M_DEBUG,"token","token line ends in whitespace. `%s'",tokline);
#endif
    
    if (*tokline == '\0' || *tokline == '#')
      continue;
    
    linecount = token_add_line(tokline,in,out,list);
    count += linecount;
    if (linecount > 0)
      numtok++;
  }
  
  fclose(tf);
  
#ifndef CAS_TOKCHECK
  if (in) {
#endif
    Msg(M_DEBUG,"token","%s: read %d for %d; %d clashes. (i%s%s)",
                 tokfile, count, numtok, clash, out?"o":"", list?"l":"");
#ifndef CAS_TOKCHECK
  } else {
    Msg(M_DEBUG,"token","%s: read %d synonyms for %d tokens. (%s%s)",
                 tokfile, count, numtok, out?"o":"", list?"l":"");
  }
#endif
}


/*
 * AddTokenLine.  Adds a line of tokens to the internal structures, which
 *                dependingon the in,out,list parameters
 */

int token_add_line(char *tokline,int in,int out,int list)
{
  long toknum=0;
  int count=0;
  int tflag;
  char *token=NULL;
  char *duptoken=NULL;

  token = strtok(tokline, "|");
  if (token == NULL) {
#ifdef CAS_TOKCHECK
    Msg(M_WARN,"token","invalid token line: %s",tokline);
#endif
    return 0;
  }

  toknum = strtol(token,NULL,0);          /* should this be base 10 ? */
  if (toknum < 0 || toknum > TOKEN_MAX) {
#ifdef CAS_TOKCHECK
    Msg(M_WARN,"token","invalid code: %ld from `%s'",toknum,token);
#endif
    return 0;
  }
  
  while ((token=strtok(NULL,"|")) != NULL) {
    tflag = token_get_flags(&token);

    /* if `-' flag set, then skip if a model is not specified */
#if 0   /* why is this commented out? */
    if ((tflag & MDL_NOTGEN) && (flags.model==MDL_UNKNOWN)) continue;
#endif

    /* if not appropriate for this moel, then skip */
    if ((tflag & flags.model) == 0) continue;

    /* copy the token into token memory, expanding escape characters */
    duptoken = token_expand(token);

    if (tflag & TF_TEXT) {
      /* only add text desciptions for output, and if a desciption
       * not already supplied
       */
      if (out && toklis->text[toknum] == NULL) {
        toklis->text[toknum] = duptoken;
      }
      continue;
    }

    /* add to `in' tokens (hash table) */
    if (in && (tflag & TF_LISTONLY) == 0) {
      token_add_to_hash(duptoken, toknum | (tflag & FLAG_MASK));
    }

    /* Add to `out' tokens (if not already specified) */
    if (out && (tflag & TF_LISTONLY) == 0) {
      if (toklis->out_nice[toknum] == NULL) {
        toklis->out_nice[toknum] = duptoken;
      }
      if (toklis->out_str[toknum] == NULL && (tflag & TF_VERBATIM) == 0) {
        toklis->out_str[toknum]=duptoken;
      }
    }

    /* Add to `list' tokens (if not already specified) */
    if (list) {
      if (toklis->list_nice[toknum] == NULL ||
         (toklis->list_nice[toknum] != NULL && (tflag & TF_LISTONLY) != 0)) {
        toklis->list_nice[toknum] = duptoken;
      }
      if ((tflag & TF_VERBATIM) == 0) {
        if (toklis->list_str[toknum] == NULL ||
            (toklis->list_str[toknum] != NULL && (tflag & TF_LISTONLY) != 0)) {
          toklis->list_str[toknum] = duptoken;
        }
      }
    }
    count++;
  }

  return count;
}


int token_get_flags(char **token_r)
{
  char *tf=NULL;
  char *temp=NULL;
  int tflag;

  if (**token_r != '{')
    return TF_DEFAULT;

  tf = (*token_r) + 1;
  for( ; ; (*token_r)++) {
    if (**token_r == '\0') {
#ifdef CAS_TOKCHECK
      Msg(M_DEBUG,"token","unclosed { in `%s'",*token_r);
#endif
      break;
    }
    if (**token_r == '}') {
      (*token_r)++;
      break;
    }
  }

  tflag = TF_DEFAULT;
  for (temp=tf; *temp != '\0' && *temp != '}'; temp++) {
    if (isdigit(*temp))
      tflag=0;
  }

  for (temp=tf; *temp != '}' && *temp != '\0'; temp++)
    switch (tolower(*temp)) {
      case '-': tflag |= MDL_NOTGEN; break;
      case '7': tflag |= MDL_7700; break;
      case '9': tflag |= MDL_9700; break;
      case '8': tflag |= MDL_9800; break;
      case '5': tflag |= MDL_9850; break;
      case 'l': tflag |= TF_LISTONLY; break;
      case 'v': tflag |= TF_VERBATIM; break;
      case 'n': tflag |= TF_NNE; break;
      case 'f': tflag |= TF_FILENAME; break;
      case 't': tflag |= TF_TEXT; break;
      default:
#ifdef CAS_TOKCHECK
        Msg(M_DEBUG,"token","Invalid token flag: `%c'",*temp)
#endif
        ;
    }

  return tflag;
}


/*
 * Duptoken.  Duplicates a token, expanding control characters (@ prefixed)
 *            as it goes.  Returns pointer to new string
 */

char *token_expand(char *token)
{
  char *i = NULL;
  char *j = NULL;
  char *duptoken = NULL;
  long num;
  char ntmp[4];
  char base = 'd';

  duptoken = toklis->tokpos;
  if (toklis->tokpos + strlen(token) > toklis->tokend) {
    Msg(EXIT_BUG,"token","insufficient memory for tokens");
  }

  for (i = token, j = duptoken; *i; i++) {
    if (!isprint((char) *i)) { 
      /* Allow `extended ascii characters' -- to make CP437.tok readable */
#ifdef CAS_DOS
      if ((unsigned char) (*i) < 128) {
#else
      if (TRUE) {
#endif
#ifdef CAS_TOKCHECK
        if (*i != '\r')
          Msg(M_DEBUG,"token","Skipping unprintable: %u",(unsigned char) *i);
				/* We don't want a warning under Unix if */
				/* there's a CR at the end of every line! */
#endif
        continue;
      }
    }
    if (*i == '@') {
      if (*++i == '\\') {
        switch (*++i) {
          case 'n': *(j++)='\n';break; /* newline */
          case 't': *(j++)='\t';break; /* tab */
          case '@': *(j++)='@';break;  /* Token file control chars */
          case '$': *(j++)='$';break;  /* obsolete but still supported */
          case '{': *(j++)='{';break;
          case 'p': *(j++)='|';break;
          case '0': *(j++)='\0';break; /* generally for blank tokens */
                                       /* could also be used for comments? */
          default:
#ifdef CAS_TOKCHECK
            Msg(M_DEBUG,"token","incorrect character code `@\\%c' in %s",*i,token)
#endif
          ;
        }
      } else {		/* @ but no backslash */
        if (*i == 'h' || *i == 'x' || *i == 'o' || *i == 'd')
          base = *(i++);
        strncpy(ntmp, i, 3);
        ntmp[3] = '\0';
        /* we add strlen(ntmp) - 1 as i is incremented at the end of the
         * for loop too.
         */
        i += strlen(ntmp) - 1;
#ifdef CAS_TOKCHECK
        if (strlen(ntmp) != 3 || isxdigit(*ntmp) == 0) {
          Msg(M_DEBUG,"token","incorrectly formatted @ value `%s' (len=%d) in `%s' from (%s)"
                            ,ntmp,strlen(ntmp),token,i);
        }
#endif
        num = strtol(ntmp, NULL, (base=='o') ? 8 : (base=='d') ? 10 : 16);
#ifdef CAS_TOKCHECK
        if (num > 255 || num < 1 ) {
          Msg(M_DEBUG,"token","@ value out of range (%ld) from %s in %s",num,ntmp,token);
        }         
#endif
        *(j++) = (unsigned char) num;
      }
    } else {  /* Normal, everyday character... */
      *(j++) = *i;
    }
  }
  *j = '\0';
  toklis->tokpos = j + 1;

  return duptoken;
}

/*
 * AddToHash.  Adds token to Hash table (toklis->in).  Returns hash value
 */

unsigned long token_add_to_hash(char *token,int toknum)
{
  unsigned long hashval;
  unsigned long hashval2;
  int probenum=0;

  hashval = Hash(token,TOKEN_HASHSIZE);
  hashval2 = Hash2(token,TOKEN_HASHSIZE_2);

  if (hashval2 == 0) {
    Msg(M_DEBUG,"token.hash","hash2 value is 0 reading `%s'. setting to %ld",
      token, TOKEN_HASHSIZE_2 / 2);
    hashval2 = TOKEN_HASHSIZE_2 / 2;
  }

  while (toklis->in[hashval].code) {
    if (!strcmp(token, toklis->in[hashval].token)) {
      if ((toklis->in[hashval].code & TOKEN_MASK) == (toknum & TOKEN_MASK)) {
#ifdef CAS_TOKCHECK
        Msg(M_DEBUG,"token","duplicate token `%s' for same code %d.",token,toknum&TOKEN_MASK);
#endif
        return 0;
      } else {
#ifdef CAS_TOKCHECK
          Msg(M_DEBUG,"token",
            "duplicate token `%s', previous code %d, ignoring new code %d.",
            token,toklis->in[hashval].code&TOKEN_MASK,toknum&TOKEN_MASK);
#endif
        return 0;
      }
    }
#ifdef CAS_TOKCHECK
      Msg(M_DEBUG,"token/hash","hash clash with `%s' adding `%s'",
                            toklis->in[hashval].token,token);
#endif
    ++clash;
    ++probenum;
    hashval += hashval2;
    hashval %= TOKEN_HASHSIZE;
  }
#ifdef CAS_TOKCHECK
  if (probenum > 1)
    Msg(M_DEBUG,"token/hash","Token `%s' required %d probes",token,probenum);
#endif
  toklis->in[hashval].code = toknum;
  toklis->in[hashval].token = token;

  return hashval;
}


/*
 * FreeTokens.  Frees memory used by token lists
 */

void token_free(void)
{
  if (toklis) {
    if (toklis->tokblock) free(toklis->tokblock);
    free(toklis);
  }
}


/*
 * CodeToToken.  Returns the token for a ctoken.  If display is set returns
 *               display token (list),  else returns program token (out)
 */

char *code_to_token(int ctoken,int which)
{
  static char unknown[]="@d\0\0\0\0";
  
  switch (which) {
    case TOKEN_LIST_NICE:
      if (toklis->list_nice[ctoken] != NULL)
        return toklis->list_nice[ctoken];
    case TOKEN_LIST_STR:
      if (toklis->list_str[ctoken] != NULL)
        return toklis->list_str[ctoken];
      break;
    case TOKEN_OUT_NICE:
      if (toklis->out_nice[ctoken] != NULL)
        return toklis->out_nice[ctoken];
    case TOKEN_OUT_STR:
      if (toklis->out_str[ctoken] != NULL)
        return toklis->out_str[ctoken];
      break;
    default:
      Msg(EXIT_BUG,"code_to_token","evil witch");
  }
  
  /* since snprintf is not necessarily available, take the modulus 1000
   * to ensure that ctoken <= 3 digits long */
   
  /* are we allowed to write to string constants, btw?  I think not */ 
  sprintf(unknown + 2, "%03u", ctoken % 1000);
  return unknown;
}


/*
 * TokenToCode.  Returns the ctoken for a string using the hash table
 */

int token_to_code(char *token,unsigned set,unsigned unset)
{
  unsigned long hashval;
  unsigned long hashval2;
  long num;
  char *check=NULL;
  unsigned flgs;

/* @bnnn converts a number to an ascii value.  Optional b sets base. */

  if (strlen(token) > 1 && token[0] == '@') {
    if (isdigit(token[1])) {
      if (strlen(token) == 4) {
        num = strtol(token + 1, &check, 10);
        if (*check == '\0') {
          if (num > TOKEN_MAX || num < 0)
            Msg(M_ERROR,"parse","@ value out of range (%ld) in `%s'",num,token);
          return ((int) num);
        } else {
          Msg(M_ERROR,"parse","error in @ value in `%s'",token);
        }
      }
    } else {
      switch (token[1]){
        case 'x': case 'h': case 'o': case 'd':
          if (strlen(token) == 5) {
            num = strtol(token + 2, &check,
                         (token[1] == 'o') ? 8 : (token[1]=='d') ? 10 : 16);
            if (!*check) {
              if (num > TOKEN_MAX || num <= 0)
                Msg(M_ERROR,"parse","@ value out of range (%ld) in %s",num,token);
              return ((int) num);
            } else {
              Msg(M_ERROR,"parse","error in @ value in `%s'",token);
            }
          }
          /* case '@': return 255; obsolete? */
        default:;
      }
    }
  }

  hashval = Hash(token,TOKEN_HASHSIZE);
  hashval2 = Hash2(token, TOKEN_HASHSIZE_2);
  if (hashval2 == 0) {
    Msg(M_DEBUG,"token.hash","hash2 value is 0 reading `%s'.  setting to %ld",
        token, TOKEN_HASHSIZE_2 / 2);
    hashval2 = TOKEN_HASHSIZE_2 / 2;
  }

  while (toklis->in[hashval].token &&
         strcmp(token, toklis->in[hashval].token)) {
    hashval += hashval2;
    hashval %= TOKEN_HASHSIZE;
  }

  if (toklis->in[hashval].code & TOKEN_MASK) {
    flgs = toklis->in[hashval].code & FLAG_MASK;
    if ((set & flgs) != set) {
      Msg(M_DEBUG,"parse","rejecting `%s' as flags 0x%x not set in 0x%x",token,set,flgs);
      return 0;
    }
    if ((unset & flgs) != 0) {
      Msg(M_DEBUG,"parse","rejecting `%s' as flags 0x%x are set in 0x%x",token,unset,flgs);
      return 0;
    }
  }
  return (toklis->in[hashval].code & TOKEN_MASK);
}

/*
 * TokensToText.  Returns string where ctokens converted to 1 byte tokens
 */

char *code_to_text(unsigned char *toks, int display)
{
  static char text[64];

  text[0] = '\0';
  if (!toks) return NULL;

  for( ; *toks != 0 && *toks != 0xff; toks++) {
    strcat(text, code_to_token(*toks, display));
  }

  return text;
}


unsigned char *text_to_code(char *text, int max)
{
  static unsigned char *toks=NULL;
  char *token=NULL;
  char *tpos=NULL;
  unsigned ctok=0;
  int n;

  toks = tsys_realloc(toks, max + 1);
  toks[max] = '\0';  /* hack so that token list can be printed */
  
  for(n=0; n<max; n++)
    toks[n] = '\xff';

  if (!text) return toks;

  token = tsys_calloc(PARSE_TOKSIZE + 1);

  n=0;
  while (*text) {
    strncpy(token, text, PARSE_TOKSIZE);
    tpos = token + strlen(token);
    while (tpos != token) {
      if ((ctok = token_to_code(token, TF_FILENAME, 0)) != 0)
        break;
      *(--tpos) = '\0';
    }
    if (ctok == 0) {
      Msg(M_WARN,"filename","Unknown or illegal token or part of token: `%c'",*text);
      text++;
    } else {
      if (n == max) {
        Msg(M_WARN,"filename","Filename too long (max: %d)",max);
        break;
      }
      toks[n++] = ctok & 0xff;
      text += (tpos - token);
    }
  }

  free(token);
  return toks;
}

/* end of token.c */
